@@ -474,7 +474,7 @@ static void log_xlate_error(ap_filter_t *f, apr_status_t rv)
     charset_filter_ctx_t *ctx = f->ctx;
     const char *msg;
     char msgbuf[100];
-    int cur;
+    apr_size_t len;
 
     switch(ctx->ees) {
     case EES_LIMIT:
@@ -492,12 +492,14 @@ static void log_xlate_error(ap_filter_t *f, apr_status_t rv)
     case EES_INCOMPLETE_CHAR:
         rv = 0;
         strcpy(msgbuf, APLOGNO(02196) "xlate filter - incomplete char at end of input - ");
-        cur = 0;
-        while ((apr_size_t)cur < ctx->saved) {
-            apr_snprintf(msgbuf + strlen(msgbuf), sizeof(msgbuf) - strlen(msgbuf),
-                         "%02X", (unsigned)ctx->buf[cur]);
-            ++cur;
-        }
+        len = ctx->saved;
+
+        /* We must ensure not to process more than what would fit in the
+         * remaining of the destination buffer, including terminating NULL */
+        if (len > (sizeof(msgbuf) - strlen(msgbuf) - 1) / 2)
+            len = (sizeof(msgbuf) - strlen(msgbuf) - 1) / 2;
+
+        ap_bin2hex(ctx->buf, len, msgbuf + strlen(msgbuf));
         msg = msgbuf;
         break;
     case EES_DOWNSTREAM:
@@ -982,6 +984,11 @@ static apr_status_t xlate_in_filter(ap_filter_t *f, apr_bucket_brigade *bb,
     apr_size_t buffer_size;
     int hit_eos;
 
+    /* just get out of the way of things we don't want. */
+    if (mode != AP_MODE_READBYTES) {
+        return ap_get_brigade(f->next, bb, mode, block, readbytes);
+    }
+
     if (!ctx) {
         /* this is SetInputFilter path; grab the preallocated context,
          * if any; note that if we decided not to do anything in an earlier
@@ -1019,7 +1026,7 @@ static apr_status_t xlate_in_filter(ap_filter_t *f, apr_bucket_brigade *bb,
              * Content-Length can't be unset here because that would break
              * being able to read the request body.
              * Processing of chunked request bodies is not impacted by this
-             * filter since the the length was not declared anyway.
+             * filter since the length was not declared anyway.
              */
             ap_log_rerror(APLOG_MARK, APLOG_TRACE1, 0, f->r,
                           "Request body length may change, resulting in "
